@@ -11,6 +11,7 @@ import smart_open.s3
 import shutil
 import tempfile
 import io
+from urllib.parse import urlencode
 
 from .__util_versioned_ref import VersionedRef
 from .__util_metadata import FIELDS
@@ -123,7 +124,7 @@ class Datalake:
             self.s3 = boto3.client('s3', region_name=AWS_REGION, config=Config(signature_version='s3v4'))
             self.resource_kwargs = None
 
-    def get_file_meta(self, file):
+    def get_s3_head(self, file):
         bucket = file['bucket']
         file_key = file['fileKey']
         if 'version' in file:
@@ -132,6 +133,10 @@ class Datalake:
         else:
             head = self.s3.head_object(Bucket = bucket, Key = file_key)
             
+        return head
+
+    def get_file_meta(self, file):
+        head = self.get_s3_head(file)
         return lowerMetadataKeys(head.get('Metadata'))
 
     def read_file(self, file, form='body'):
@@ -263,6 +268,72 @@ class Datalake:
             'version': response.get('VersionId', '')
         }
 
+    def update_metadata_tags(self, file, custom_meta, custom_tags):
+        bucket = file['bucket']
+        file_key = file['fileKey']
+
+        head = self.get_s3_head(file)
+        current_meta = lowerMetadataKeys(head.get('Metadata'))
+
+        if not FIELDS['FILE_ID'] in current_meta:
+            raise Exception('no FILE_ID in meta!')
+
+        if (not custom_meta) and (not custom_tags):
+            raise Exception('No metadata or tags provided')
+
+        isASCII = lambda s: s and isinstance(s, str) and bool(re.match(r'^[\x00-\x7F]*$', s))
+        
+        custom_meta_str = current_meta.get(FIELDS['CUSTOM_METADATA'], '') or ''
+        current_custom_meta = query_string.parse(custom_meta_str)
+        if custom_meta:
+            custom_meta_merged = {**current_custom_meta, **custom_meta}
+            custom_meta_merged = {k:v for k,v in custom_meta_merged.items() if v is not None}
+            for k,v in custom_meta_merged.items():
+                if not isASCII(k):
+                    raise Exception(f'Metadata key {k} contains non-ASCII character') 
+                if not isASCII(v):
+                    raise Exception(f'Metadata value {v} contains non-ASCII character')
+            custom_meta_str = urlencode(custom_meta_merged)
+
+        custom_tags_str = current_meta.get(FIELDS['CUSTOM_TAGS'], '')
+        if custom_tags:
+            for t in custom_tags:
+                if not isASCII(t):
+                    raise Exception(f'Tag {t} contains non-ASCII character') 
+            custom_tags_str = ','.join(custom_tags)
+
+        params = {
+            'Bucket': bucket,
+            'CopySource': f'/{bucket}/{file_key}',
+            'CopySourceIfUnmodifiedSince': head['LastModified'], # ensure no conflict
+            'Key': file_key,
+            'ContentEncoding': head.get('ContentEncoding', None),
+            'ContentType': head['ContentType'],
+            'Metadata': {
+                **current_meta,
+                FIELDS['CUSTOM_METADATA']: custom_meta_str,
+                FIELDS['CUSTOM_TAGS']: custom_tags_str
+            },
+            'MetadataDirective': 'REPLACE',
+            'ServerSideEncryption': 'aws:kms',
+            'SSEKMSKeyId': head.get('SSEKMSKeyId', None),
+        }
+
+        if len(custom_meta_str) + len(custom_tags_str) >= 1024 * 1.5:
+            raise Exception('Metadata and tags length larger than 1.5KB') 
+
+        params = {k:v for k,v in params.items() if v is not None}
+        response = self.s3.copy_object(**params)
+
+        return {
+            'type': 's3file',
+            'bucket': bucket,
+            'fileKey': file_key,
+            'fileId': current_meta[FIELDS['FILE_ID']],
+            # fakeS3 does not return VersionId, so use '' to avoid an exception
+            'version': response.get('VersionId', '')
+        }
+
     def write_ids(self, context, content_obj, file_suffix, raw_file, file_meta, ids, source_type):
         ids_obj = VersionedRef(composite=ids)
         file_name = f'{ids_obj.name}_{ids_obj.version}_{file_suffix}'
